# -*- coding: utf-8 -*-

FAILURE_CLASSES = ("not classified", "new failure not classified", "fixed by commit")
INTERMITTENT_CLASSES = (
    "intermittent",
    "autoclassified intermittent",
    "intermittent needs bugid",
)
TASK_FINAL_STATES = ("completed", "failed", "exception")
